/*****************************************************************
 * Copyright (c) ParaSoft Corporation 1991.  All rights reserved.*
 * THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF ParaSoft       *
 * The copyright notice above does not evidence any              *
 * actual or intended publication of such source code.           *
 *****************************************************************/

#include "extern.h"

static flag_case_sensitive = 0;

int table_li;

T_Entry *table_scope[MAX_TABLE_LEVEL][HASHSIZE];   /* local scope for procedures */
T_Entry *table_intrin [HASHSIZE];

int table_level = 1;

static char *str_upper (str)  /* R. string converted to upper, cannot be call recur. */
char *str;
{
	register char *f = str, *t;
	static char *new = (char*) 0;

	if (new != (char*) 0)
		U_free (new);

	new = (char*) U_malloc (strlen (str) + 1);
	t = new;

	if (f == (char*) 0)
		return str;

	while (*f != '\0')
		if ('a' <= *f && *f <= 'z')
			*t++ = (*f++) - 'a' + 'A';
		else
			*t++ = *f++;

	return new;
}

table_init ()  /* initializes table */
{
	int l, i;

	for (l = 0; l < MAX_TABLE_LEVEL; l++)
		for (i = 0; i < HASHSIZE; i++)
			table_scope[l][i] = (T_Entry*) 0;
}

unsigned hash (arg_name)
char *arg_name;
{
	unsigned register hashval;
	register char *name = arg_name;

	for (hashval = 0; *name != '\0'; name++)
		hashval = *name + 31 * hashval;

	return hashval % HASHSIZE;
}

T_Entry *lookup_all_glo_bef (name)
char *name;
{
	T_Entry *entry;

	if ((entry = lookup_glo (name)) == TN)
		return lookup_loc (name);
	else
		return entry;
}

T_Entry *lookup_all (name)
char *name;
{
	T_Entry *entry;

	if ((entry = lookup_loc (name)) == TN)
		return lookup_glo (name);
	else
		return entry;
}

T_Entry *lookup_loc (name)
char *name;
{
	return lookup (name, TAB_LOCAL);
}

T_Entry *lookup_user (name)
char *name;
{
	return lookup (name, TAB_GLOBAL);
}

T_Entry *lookup_glo (name)
char *name;
{
	T_Entry *entry, *loc_entry = lookup_loc (name);

	if (loc_entry == TN) {
		if ((entry = lookup_intrin (name)) != TN)
			return entry;
		else
			return lookup (name, TAB_GLOBAL);
	}

	if (IS_external (loc_entry)) {
		return lookup (name, TAB_GLOBAL);
	}
	else if (IS_intrinsic (loc_entry)) {
		if ((entry = lookup_intrin (name)) == TN) {
			IS_intrinsic (loc_entry) = 0;  /* to avoid repetition of error message */
			sprintf (serr, "there is no supported intrinsic called '%s'", name);
			warning (serr);

			entry = lookup (name, TAB_GLOBAL);
		}

		return entry;
	}
	else {  /* first looks at intrinsic and then at user */
		if ((entry = lookup_intrin (name)) != TN)
			return entry;
		else
			return lookup (name, TAB_GLOBAL);
	}
}

T_Entry *lookup_intrin (name)
char *name;
{
	T_Entry *entry;

	name = str_upper (name);

	for (entry = table_intrin [hash (name)]; entry != (T_Entry*) 0; entry = entry->next)
		if (strcmp (name, entry->name) == 0)
			return entry;

	return (T_Entry*) 0;
}

T_Entry *lookup (name, iscope)
char *name;
int iscope;
{
	T_Entry *entry;

	name = str_upper (name);

	if (iscope == TAB_LOCAL) {
		for (entry = table_scope[table_level][hash (name)]; entry != (T_Entry*) 0; entry = entry->next)
			if (strcmp (name, entry->name) == 0)
				return entry;
	}
	else if (iscope == TAB_GLOBAL) {
		for (entry = table_scope[0][hash (name)]; entry != (T_Entry*) 0; entry = entry->next)
			if (strcmp (name, entry->name) == 0)
				return entry;
	}
#ifdef CHECK
	else
		SYSERR;
#endif

	return (T_Entry*) 0;
}

T_Entry *install_loc (name, what, point)
char *name;
int what;
char *point;
{
	return install (name, what, point, TAB_LOCAL);
}
T_Entry *install_glo (name, what, point)
char *name;
int what;
char *point;
{
	return install (name, what, point, TAB_GLOBAL);
}

T_Entry *install (name, what, point, iscope)
char *name;
int what;
char *point;
int iscope;
{
	T_Entry *entry;
	unsigned hashval;
	char *Uname;  /* Uname is unique name */

	if ((entry = lookup (name, iscope)) != (T_Entry*) 0) {
		if (what == TAB_NAME && !IS_intrinsic (entry)) {
			sprintf (serr, "identifier '%s' already defined", name);
			fatal (serr);
		}
	}

	if (entry == TN) {
		Uname = U_str_copy_upper (name);

		hashval = hash (Uname);
		entry = (T_Entry*) U_malloc (sizeof (T_Entry));
		entry->name = Uname;
		entry->orig_name = name;

		if (iscope == TAB_GLOBAL) {
			entry->next = table_scope[0][hashval];
			table_scope[0][hashval] = entry;
		}
		else if (iscope == TAB_LOCAL) {
			entry->next = table_scope[table_level][hashval];
			table_scope[table_level][hashval] = entry;
		}
#ifdef CHECK
		else
			SYSERR;
#endif
	}

	switch (what) {

	case TAB_INSTALL:
	case TAB_NAME:
		break;

	case TAB_NODE:
		if (entry->node == NN)  
			entry->node = (Node*) point;
		break;
/* ##BEG##, 9/26/91, SHU */
	case TAB_DIST:
		if (entry->dist == NN)  /* sets only the first node */
			entry->dist = (Node*) point;
		break;
/* ##END## */
/* the program that calls install should assume that typ is gone, so install should
   deallocate typ if it doesn't use it */

	case TAB_TYP:
		if (entry->typ != (Typ*) 0) {
			sprintf (serr, "typ for '%s' already set", name);
			syserr (serr);
		}
		entry->typ = (Typ*) point;
		break;

	case TAB_TYP_ETYPE:
		{
		Etype *etype = (Etype*) point;

#ifdef CHECK
		if (etype == EN)
			SYSERR;
#endif

		if (entry->typ == (Typ*) 0)
			entry->typ = typ_malloc ();

		if (entry->typ->etype == EN) {
			entry->typ->etype = etype;
		}
		else {
			if (etype_ne (entry->typ->etype, etype)) {
				if (IS_implicit_type (entry)) {
					sprintf (serr
					, "redeclaration of '%s' with incompatible type: was implicit %s, is %s"
					, name, etype_name (entry->typ->etype), etype_name(etype));
				}
				else {
					sprintf (serr
					, "redeclaration of '%s' with incompatible type: was %s, is %s"
					, name, etype_name (entry->typ->etype), etype_name(etype));
				}

				if (iscope == TAB_LOCAL) {
					if ((IS_external (entry) || IS_intrinsic (entry))
					&& IS_implicit_type (entry))
						;
					else
						fatal (serr);
				}
				else if (!IS_intrinsic (entry))  /* it is OK to override intrinsics */
					warning (serr);  /* gives just warning for functions */
			}
			else {
				if (iscope == TAB_LOCAL) {
					if (IS_implicit_type (entry)) {
						IS_implicit_type (entry) = 0;
					}
					else {
/* ##BEGIN##, 02/23/1992, ZEKI */
/* output_of_intrinsic array is in sybbol table before
   but I am putting again to create decleration 
   according lhs (I know temproarary size after asignment statement */

                                            if(!IS_output_of_intrinsic(entry)){
						sprintf (serr
						, "identifier '%s' is declared more than once with the same type"
						, name);
						warning (serr);
					      }
					}
				}
			}
		}
		}
		break;

	case TAB_TYP_SHAPE:
		{
		Shape *shape = (Shape*) point;

		if (entry->typ == (Typ*) 0) {
			entry->typ = typ_malloc ();
			entry->typ->shape = shape;
		}
		else {
			if (entry->typ->shape != (Shape*) 0
			&& !typ_shape_eq (entry->typ->shape, shape)) {
		/* ZEKI 	sprintf (serr, "redimensioning of '%s'", name);
				fatal (serr); */
			}
			else {
				entry->typ->shape = shape;
			}
		}
		}
		break;

	case TAB_INTERF:
		if (entry->interf != (Interf*) 0) {
			sprintf (serr, "interface for '%s' already defined", name);
			fatal (serr);
		}
		entry->interf = (Interf*) point;
		break;

	case TAB_COMPILE_TIME:
		if (entry->compile_time != (PtF) 0) {
			sprintf (serr, "compile-time for '%s' already set", name);
			syserr (serr);
		}
		entry->interf = (Interf*) point;
		break;

	case TAB_LABEL:
		SET_label (entry);
		break;

#ifdef CHECK
	default:
		SYSERR;
		break;
#endif
	}

	return entry;
}

#include "prot_mask.h"
int prot_mask = PROT_MASK;

#ifdef CHECK
table_show_terse ()
{
	T_Entry *entry;

	for_TABLE_LOOP (entry)
		printf ("%s\n", entry->orig_name);
}

table_show_entry (entry)
T_Entry *entry;
{
	printf ("%s: ", entry->orig_name);
	printf ("(%s) ", entry->tempname);
	typ_show (entry->typ);
	printf (" ");
	print_attr (entry);
	printf (" ");
	interf_show (entry->interf);
	over_show (entry->over);
}

table_show1 ()
{
	table_show1_global ();
	table_show1_local ();
}

table_show1_global ()
{
	int i;
	T_Entry *entry;

	printf ("table_show:--- G L O B A L ------------------------------------\n");
	for (i = 0; i < HASHSIZE; i++)
		for (entry = table_scope[0][i]; entry != (T_Entry*) 0; entry = entry->next) {
			printf ("************************************************************\n");
			table_show_entry (entry);
			printf ("\n************************************************************\n");
		}
}

table_show1_local ()
{
	int i;
	T_Entry *entry;

	printf ("table_show:---  L O C A L --------------------------------------\n");
	for (i = 0; i < HASHSIZE; i++)
		for (entry = table_scope[table_level][i]; entry != (T_Entry*) 0;
		entry = entry->next) {
			table_show_entry (entry);
			putchar ('\n');
		}

	printf ("END table_show:----------------------------------------------\n");
}

static print_attr (entry)
T_Entry *entry;
{
	printf ("[ ");
	if (IS_optional (entry))
		printf ("OPTIONAL ");
	if (IS_assumed_shape (entry))
		printf ("ASSUMED_SHAPE ");
	if (IS_assumed_pass (entry))
		printf ("ASSUMED_PASS ");
	if (IS_array (entry))
		printf ("ARRAY ");
	if (IS_compile_time (entry))
		printf ("COMPILE_TIME ");
	if (IS_distributed (entry))
		printf ("DISTRIBUTED ");
/* ##BEG##, 9/26/91, SHU */
	if (IS_decomposed (entry))
		printf ("DECOMPOSED ");
/* ##END## */

/* ##BEG##, 10/27/91, ZEKI */
	if (IS_aligned (entry))
		printf ("ALIGNED ");
/* ##END## */
	if (IS_intrinsic (entry))
		printf ("INTRINSIC ");
	if (IS_argument (entry))
		printf ("ARGUMENT ");
	if (IS_automatic (entry))
		printf ("AUTOMATIC ");
	if (IS_parameter (entry))
		printf ("PARAMETER ");
	if (IS_function (entry))
		printf ("FUNCTION ");
	if (IS_subroutine (entry))
		printf ("SUBROUTINE ");
	if (IS_elemental (entry))
		printf ("ELEMENTAL ");
	if (IS_external (entry))
		printf ("EXTERNAL ");
	if (IS_implicit (entry))
		printf ("IMPLICIT ");
	if (IS_explicit (entry))
		printf ("EXPLICIT ");
	if (IS_defined_proc (entry))
		printf ("DEFINED_PROC ");
	if (IS_temporary (entry))
		printf ("TEMPORARY ");
	if (IS_alloc_temp (entry))
		printf ("ALLOC_TEMP ");
	if (IS_result (entry))
		printf ("RESULT ");
	if (IS_over (entry))
		printf ("OVER ");
	if (IS_common (entry))
		printf ("COMMON ");
	if (IS_save (entry))
		printf ("SAVE ");
	if (IS_has_initexpr (entry))
		printf ("HAS_INITEXPR ");
	if (IS_blank_common (entry))
		printf ("BLANK_COMMON ");
	if (IS_named_common (entry))
		printf ("NAMED_COMMON ");
	if (IS_no_checking (entry))
		printf ("NO_CHECKING ");
	if (IS_constant (entry))
		printf ("CONSTANT ");
	if (IS_label (entry))
		printf ("LABEL ");
	if (IS_label_ref (entry))
		printf ("LABEL_REF ");
	if (IS_label_def (entry))
		printf ("LABEL_DEF ");
	if (IS_label_format (entry))
		printf ("LABEL_FORMAT");
	if (IS_label_io (entry))
		printf ("LABEL_IO");
	if (IS_label_temp (entry))
		printf ("LABEL_TEMP");
	if (IS_namelist (entry))
		printf ("NAMELIST");
	if (IS_intent_in (entry))
		printf ("IN");
	if (IS_intent_out(entry))
		printf ("OUT");
	printf ("]");
}
#endif

table_end ()  /* clears local table */
{
	int i;
	T_Entry *old, *entry;

	for (i = 0; i < HASHSIZE; i++) {
		entry = table_scope [table_level][i];
		while (entry != (T_Entry*) 0) {
			old = entry;
			entry = entry->next;
			U_free (old);
		}
		table_scope[table_level][i] = (T_Entry*) 0;  /* should free chain also */
	}
}

T_Entry *table_entry_cp (entry)  /* sort of returns a copy of the entry */
T_Entry *entry;
{
	T_Entry *nentry;

	nentry = (T_Entry*) U_malloc (sizeof (T_Entry));
	memcpy (nentry, entry, sizeof (T_Entry));

	return nentry;
}

push_scope ()
{
	if (++table_level >= MAX_TABLE_LEVEL)
		SYSERR_TERM;
}

pop_scope ()
{
	if (--table_level <= 0)
		SYSERR_TERM;
}

table_user_to_intrin ()
{
	int i;

	for (i = 0; i < HASHSIZE; i++) {
		table_intrin [i] = table_scope [0][i];
		table_scope [0][i] = TN;
	}
}
